@@ -26,15 +26,19 @@
 
 #include "htmlclue.h"
 #include "htmlclueflow.h"
+#include "htmlcluealigned.h"
 #include "htmlcluev.h"
 #include "htmlcolor.h"
 #include "htmlcolorset.h"
 #include "htmlcursor.h"
 #include "htmlengine.h"
 #include "htmlengine-edit.h"
+#include "htmlengine-save.h"
+#include "htmlframe.h"
 #include "htmlinterval.h"
 #include "htmlobject.h"
 #include "htmlpainter.h"
+#include "htmltable.h"
 #include "htmltext.h"
 #include "htmlrule.h"
 #include "htmltype.h"
@@ -49,28 +53,23 @@ HTMLObjectClass html_object_class;
 
 /* HTMLObject virtual methods.  */
 
-static void
-free_data (GQuark id, gpointer data, gpointer user_data)
-{
-	g_free (data);
-}
-
 static void
 destroy (HTMLObject *self)
 {
+#define GTKHTML_MEM_DEBUG 1
+#if GTKHTML_MEM_DEBUG
+	self->parent = HTML_OBJECT (0xdeadbeef);
+	self->next = HTML_OBJECT (0xdeadbeef);
+	self->prev = HTML_OBJECT (0xdeadbeef);
+#else
+	self->next = NULL;
+	self->prev = NULL;
+#endif
+	g_datalist_clear (&self->object_data);
+	
 	if (self->redraw_pending) {
 		self->free_pending = TRUE;
 	} else {
-
-#if GTKHTML_MEM_DEBUG
-		self->next = 0xdeadbeef;
-		self->prev = 0xdeadbeef;
-#else
-		self->next = NULL;
-		self->prev = NULL;
-#endif
-		g_datalist_foreach (&self->object_data, free_data, NULL);
-		g_datalist_clear (&self->object_data);
 		g_free (self);
 	}
 }
@@ -93,7 +92,7 @@ copy (HTMLObject *self,
 	dest->pref_width = self->pref_width;
 	dest->percent = self->percent;
 	dest->flags = self->flags;
-	dest->redraw_pending = self->redraw_pending;
+	dest->redraw_pending = FALSE;
 	dest->selected = self->selected;
 	dest->free_pending = FALSE;
 	dest->change = self->change;
@@ -103,7 +102,7 @@ copy (HTMLObject *self,
 }
 
 static HTMLObject *
-op_copy (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, guint *len)
+op_copy (HTMLObject *self, HTMLObject *parent, HTMLEngine *e, GList *from, GList *to, guint *len)
 {
 	if ((!from || GPOINTER_TO_INT (from->data) == 0)
 	    && (!to || GPOINTER_TO_INT (to->data) == html_object_get_length (self))) {
@@ -140,7 +139,7 @@ op_cut (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left, GL
 }
 
 static gboolean
-merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList *left, GList *right)
+merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList **left, GList **right, HTMLCursor *cursor)
 {
 	return FALSE;
 }
@@ -163,17 +162,19 @@ split (HTMLObject *self, HTMLEngine *e, HTMLObject *child, gint offset, gint lev
 	}
 
 	if (offset) {
-		if (!self->next)
+		if (!self->next) {
 			html_clue_append (HTML_CLUE (self->parent), html_engine_new_text_empty (e));
+		}
 		*left  = g_list_prepend (*left,  self);
 		*right = g_list_prepend (*right, self->next);
 	} else {
-		if (!self->prev)
-			html_clue_prepend (HTML_CLUE (self->parent), html_engine_new_text_empty (e));
+		if (!self->prev) {
+			e->cursor->object = html_engine_new_text_empty (e);
+			e->cursor->offset = 0;
+			html_clue_prepend (HTML_CLUE (self->parent), e->cursor->object);
+		}
 		*left  = g_list_prepend (*left,  self->prev);
 		*right = g_list_prepend (*right, self);
-		e->cursor->object = self->prev;
-		e->cursor->offset = html_object_get_length (self->prev);
 	}
 	level--;
 
@@ -223,44 +224,32 @@ fit_line (HTMLObject *o,
 	  HTMLPainter *painter,
 	  gboolean start_of_line,
 	  gboolean first_run,
+	  gboolean next_to_floating,
 	  gint width_left)
 {
-	html_object_calc_size (o, painter);
-	return (o->width <= width_left || first_run) ? HTML_FIT_COMPLETE : HTML_FIT_NONE;
+	html_object_calc_size (o, painter, FALSE);
+	return (o->width <= width_left || (first_run && !next_to_floating)) ? HTML_FIT_COMPLETE : HTML_FIT_NONE;
 }
 
 static gboolean
-calc_size (HTMLObject *o,
-	   HTMLPainter *painter)
+calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed_objs)
 {
 	return FALSE;
 }
 
 static gint
-calc_min_width (HTMLObject *o,
-		HTMLPainter *painter)
+calc_min_width (HTMLObject *o, HTMLPainter *painter)
 {
-	html_object_calc_size (o, painter);
+	html_object_calc_size (o, painter, FALSE);
 	return o->width;
 }
 
 static gint
-calc_preferred_width (HTMLObject *o,
-		      HTMLPainter *painter)
+calc_preferred_width (HTMLObject *o, HTMLPainter *painter)
 {
-	html_object_calc_size (o, painter);
+	html_object_calc_size (o, painter, FALSE);
 	return o->width;
 }
-
-static void
-set_max_ascent (HTMLObject *o, HTMLPainter *painter, gint a)
-{
-}
-	
-static void
-set_max_descent (HTMLObject *o, HTMLPainter *painter, gint d)
-{
-}
 	
 static void
 set_max_width (HTMLObject *o, HTMLPainter *painter, gint max_width)
@@ -268,14 +257,19 @@ set_max_width (HTMLObject *o, HTMLPainter *painter, gint max_width)
 	o->max_width = max_width;
 }
 
+static void
+set_max_height (HTMLObject *o, HTMLPainter *painter, gint max_height)
+{
+}
+
 static gint
-get_left_margin (HTMLObject *self, HTMLPainter *painter, gint y)
+get_left_margin (HTMLObject *self, HTMLPainter *painter, gint y, gboolean with_aligned)
 {
 	return 0;
 }
 
 static gint
-get_right_margin (HTMLObject *self, HTMLPainter *painter, gint y)
+get_right_margin (HTMLObject *self, HTMLPainter *painter, gint y, gboolean with_aligned)
 {
 	return MAX (self->max_width, self->width);
 }
@@ -288,9 +282,9 @@ set_painter (HTMLObject *o, HTMLPainter *painter)
 static void
 reset (HTMLObject *o)
 {
-	o->width = 0;
-	o->ascent = 0;
-	o->descent = 0;
+	/* o->width = 0;
+	   o->ascent = 0;
+	   o->descent = 0; */
 }
 
 static const gchar *
@@ -385,7 +379,7 @@ relayout (HTMLObject *self,
 		self->y -= prev_ascent + prev_descent;
 	}
 
-	changed = html_object_calc_size (self, engine->painter);
+	changed = html_object_calc_size (self, engine->painter, FALSE);
 
 	if (prev_width == self->width
 	    && prev_ascent == self->ascent
@@ -594,7 +588,12 @@ tail (HTMLObject *self)
 	return NULL;
 }
 
-
+static HTMLClearType
+get_clear (HTMLObject *self)
+{
+	return HTML_CLEAR_NONE;
+}
+
 /* Class initialization.  */
 
 void
@@ -627,9 +626,8 @@ html_object_class_init (HTMLObjectClass *klass,
 	klass->is_transparent = is_transparent;
 	klass->fit_line = fit_line;
 	klass->calc_size = calc_size;
-	klass->set_max_ascent = set_max_ascent;
-	klass->set_max_descent = set_max_descent;
 	klass->set_max_width = set_max_width;
+	klass->set_max_height = set_max_height;
 	klass->get_left_margin = get_left_margin;
 	klass->get_right_margin = get_right_margin;
 	klass->set_painter = set_painter;
@@ -666,6 +664,7 @@ html_object_class_init (HTMLObjectClass *klass,
 	klass->head = head;
 	klass->tail = tail;
 	klass->get_engine = get_engine;
+	klass->get_clear = get_clear;
 }
 
 void
@@ -730,9 +729,9 @@ html_object_dup (HTMLObject *object)
 }
 
 HTMLObject *
-html_object_op_copy (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, guint *len)
+html_object_op_copy (HTMLObject *self, HTMLObject *parent, HTMLEngine *e, GList *from, GList *to, guint *len)
 {
-	return (* HO_CLASS (self)->op_copy) (self, e, from, to, len);
+	return (* HO_CLASS (self)->op_copy) (self, parent, e, from, to, len);
 }
 
 HTMLObject *
@@ -742,13 +741,13 @@ html_object_op_cut (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GLi
 }
 
 gboolean
-html_object_merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList *left, GList *right)
+html_object_merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList **left, GList **right, HTMLCursor *cursor)
 {
 	if ((HTML_OBJECT_TYPE (self) == HTML_OBJECT_TYPE (with)
 	     /* FIXME */
 	     || (HTML_OBJECT_TYPE (self) == HTML_TYPE_TABLECELL && HTML_OBJECT_TYPE (with) == HTML_TYPE_CLUEV)
 	     || (HTML_OBJECT_TYPE (with) == HTML_TYPE_TABLECELL && HTML_OBJECT_TYPE (self) == HTML_TYPE_CLUEV))
-	    && (* HO_CLASS (self)->merge) (self, with, e, left, right)) {
+	    && (* HO_CLASS (self)->merge) (self, with, e, left, right, cursor)) {
 		if (with->parent)
 			html_object_remove_child (with->parent, with);
 		html_object_destroy (with);
@@ -796,6 +795,10 @@ html_object_calc_abs_position (HTMLObject *o,
 	for (p = o->parent; p != NULL; p = p->parent) {
 		*x_return += p->x;
 		*y_return += p->y - p->ascent;
+		if (html_object_is_frame (p)) {
+			*x_return += GTK_HTML (HTML_FRAME (p)->html)->engine->leftBorder;
+			*y_return += GTK_HTML (HTML_FRAME (p)->html)->engine->topBorder;
+		}
 	}
 }
 
@@ -866,29 +869,21 @@ html_object_fit_line (HTMLObject *o,
 		      HTMLPainter *painter,
 		      gboolean start_of_line, 
 		      gboolean first_run,
+		      gboolean next_to_floating,
 		      gint width_left)
 {
-	return (* HO_CLASS (o)->fit_line) (o, painter, start_of_line,
-					   first_run, width_left);
+	return (* HO_CLASS (o)->fit_line) (o, painter, start_of_line, first_run, next_to_floating, width_left);
 }
 
 gboolean
-html_object_calc_size (HTMLObject *o,
-		       HTMLPainter *painter)
+html_object_calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed_objs)
 {
-	return (* HO_CLASS (o)->calc_size) (o, painter);
-}
+	gboolean rv;
 
-void
-html_object_set_max_ascent (HTMLObject *o, HTMLPainter *painter, gint a)
-{
-	(* HO_CLASS (o)->set_max_ascent) (o, painter, a);
-}
+	rv = (* HO_CLASS (o)->calc_size) (o, painter, changed_objs);
+	o->change &= ~HTML_CHANGE_SIZE;
 
-void
-html_object_set_max_descent (HTMLObject *o, HTMLPainter *painter, gint d)
-{
-	(* HO_CLASS (o)->set_max_descent) (o, painter, d);
+	return rv;
 }
 
 void
@@ -897,16 +892,22 @@ html_object_set_max_width (HTMLObject *o, HTMLPainter *painter, gint max_width)
 	(* HO_CLASS (o)->set_max_width) (o, painter, max_width);
 }
 
-gint
-html_object_get_left_margin (HTMLObject *self, HTMLPainter *painter, gint y)
+void
+html_object_set_max_height (HTMLObject *o, HTMLPainter *painter, gint max_height)
 {
-	return (* HO_CLASS (self)->get_left_margin) (self, painter, y);
+	(* HO_CLASS (o)->set_max_height) (o, painter, max_height);
 }
 
 gint
-html_object_get_right_margin (HTMLObject *self, HTMLPainter *painter, gint y)
+html_object_get_left_margin (HTMLObject *self, HTMLPainter *painter, gint y, gboolean with_aligned)
 {
-	return (* HO_CLASS (self)->get_right_margin) (self, painter, y);
+	return (* HO_CLASS (self)->get_left_margin) (self, painter, y, with_aligned);
+}
+
+gint
+html_object_get_right_margin (HTMLObject *self, HTMLPainter *painter, gint y, gboolean with_aligned)
+{
+	return (* HO_CLASS (self)->get_right_margin) (self, painter, y, with_aligned);
 }
 
 static void
@@ -969,6 +970,17 @@ html_object_get_target (HTMLObject *o)
 	return (* HO_CLASS (o)->get_target) (o);
 }
 
+gchar *
+html_object_get_complete_url (HTMLObject *o)
+{
+	const gchar *url, *target;
+
+	url = html_object_get_url (o);
+	target = html_object_get_target (o);
+	return url || target ? g_strconcat (url ? url : "#", url ? (target && *target ? "#" : NULL) : target,
+					      url ? target : NULL, NULL) : NULL;
+}
+
 const gchar *
 html_object_get_src (HTMLObject *o)
 {
@@ -1081,13 +1093,25 @@ html_object_append_selection_string (HTMLObject *self,
 	(* HO_CLASS (self)->append_selection_string) (self, buffer);
 }
 
-
 HTMLEngine *
 html_object_get_engine (HTMLObject *self, HTMLEngine *e)
 {
 	return (* HO_CLASS (self)->get_engine) (self, e);
 }
 
+HTMLEngine *
+html_object_engine (HTMLObject *o, HTMLEngine *e)
+{
+	while (o) {
+		e = html_object_get_engine (o, e);
+		if (html_object_is_frame (o))
+			break;
+		o = o->parent;
+	}
+
+	return e;
+}
+
 void
 html_object_forall (HTMLObject *self,
 		    HTMLEngine *e,
@@ -1119,6 +1143,19 @@ html_object_is_text (HTMLObject *object)
 	return (type == HTML_TYPE_TEXT || type == HTML_TYPE_LINKTEXT);
 }
 
+gboolean
+html_object_is_clue (HTMLObject *object)
+{
+	HTMLType type;
+
+	g_return_val_if_fail (object != NULL, FALSE);
+
+	type = HTML_OBJECT_TYPE (object);
+
+	return (type == HTML_TYPE_CLUE || type == HTML_TYPE_CLUEV || type == HTML_TYPE_TABLECELL
+		|| type == HTML_TYPE_CLUEFLOW || type == HTML_TYPE_CLUEALIGNED);
+}
+
 HTMLObject *
 html_object_next_not_type (HTMLObject *object, HTMLType t)
 {
@@ -1346,13 +1383,16 @@ html_object_cursor_forward (HTMLObject *self, HTMLCursor *cursor)
 gboolean
 html_object_cursor_backward (HTMLObject *self, HTMLCursor *cursor)
 {
+	HTMLObject *prev;
+
 	g_assert (self);
 	g_assert (cursor->object == self);
 
 	if (html_object_is_container (self))
 		return FALSE;
 
-	if (cursor->offset > 1 || (!html_object_prev_not_slave (self) && cursor->offset > 0)) {
+	if (cursor->offset > 1 || (cursor->offset > 0 && (! (prev = html_object_prev_not_slave (self))
+							  || HTML_IS_CLUEALIGNED (prev)))) {
 		cursor->offset --;
 		cursor->position --;
 		return TRUE;
@@ -1453,10 +1493,7 @@ next_object_uptree_cursor (HTMLObject *obj, HTMLObject * (*next_fn ) (HTMLObject
 static HTMLObject *
 move_object_downtree_cursor (HTMLObject *obj, HTMLObject * (*down_fn ) (HTMLObject *))
 {
-	HTMLObject *down;
-
-	while ((down = (*down_fn) (obj))) {
-		obj = down;
+	while ((obj = (*down_fn) (obj))) {
 		if (html_object_accepts_cursor (obj))
 			break;
 	}
@@ -1468,29 +1505,43 @@ static HTMLObject *
 move_object_cursor (HTMLObject *obj, gint *offset, gboolean forward,
 		    HTMLObject * (*next_fn ) (HTMLObject *), HTMLObject * (*down_fn ) (HTMLObject *))
 {
-	HTMLObject *down;
+	HTMLObject *down, *before;
 
-	if (((*offset == 0 && forward) || (*offset && !forward)) && html_object_is_container (obj))
-		if ((down = (*down_fn) (obj))) {
-			down = move_object_downtree_cursor (down, down_fn);
-			if (down) {
-				if (html_object_is_container (down))
-					*offset = forward ? 0 : 1; /* FIXME for prev it's reversed */
-				return down;
+	do {
+		gboolean found = FALSE;
+		if (((*offset == 0 && forward) || (*offset && !forward)) && html_object_is_container (obj))
+			if ((down = (*down_fn) (obj))) {
+				down = move_object_downtree_cursor (down, down_fn);
+				if (down) {
+					if (html_object_is_container (down))
+						*offset = forward ? 0 : 1;
+					return down;
+				}
 			}
-		}
 
-	obj = next_object_uptree_cursor (obj, next_fn);
-	if (obj) {
-		if (html_object_accepts_cursor (obj)) {
-			if (html_object_is_container (obj))
-				*offset = forward ? 1 : 0; /* FIXME for prev it's reversed */
-		} else {
-			obj = move_object_downtree_cursor (obj, down_fn);
-			if (html_object_is_container (obj))
-				*offset = forward ? 0 : 1; /* FIXME for prev it's reversed */
-		}
-	}
+		before = obj;
+		do {
+			obj = next_object_uptree_cursor (obj, next_fn);
+			if (obj) {
+				if (html_object_accepts_cursor (obj)) {
+					if (html_object_is_container (obj))
+						*offset = before->parent == obj->parent
+							? forward ? 0 : 1
+							: forward ? 1 : 0;
+					found = TRUE;
+				} else {
+					HTMLObject *down;
+					down = move_object_downtree_cursor (obj, down_fn);
+					if (down) {
+						if (html_object_is_container (down))
+							*offset = forward ? 0 : 1;
+						obj = down;
+						found = TRUE;
+					}
+				}
+			}
+		} while (obj && !found);
+	} while (obj && !html_object_accepts_cursor (obj));
 
 	return obj;
 }
@@ -1524,7 +1575,7 @@ html_object_get_index (HTMLObject *self, guint offset)
 void
 html_object_set_data (HTMLObject *object, const gchar *key, const gchar *value)
 {
-	g_datalist_set_data (&object->object_data, key, g_strdup (value));
+	g_datalist_set_data_full (&object->object_data, key, g_strdup (value), g_free);
 }
 
 gpointer
@@ -1538,7 +1589,10 @@ copy_data (GQuark key_id, gpointer data, gpointer user_data)
 {
 	HTMLObject *o = HTML_OBJECT (user_data);
 
-	g_datalist_id_set_data (&o->object_data, key_id, g_strdup ((gchar *)data));
+	/* printf ("copy key: '%s'\n", g_quark_to_string (key_id)); */
+	g_datalist_set_data_full (&o->object_data,
+				  g_strdup (g_quark_to_string (key_id)),
+				  g_strdup ((gchar *) data), g_free);
 }
 
 void
@@ -1547,6 +1601,71 @@ html_object_copy_data_from_object (HTMLObject *dst, HTMLObject *src)
 	g_datalist_foreach (&src->object_data, copy_data, dst);
 }
 
+static void
+object_save_data (GQuark key_id, gpointer data, gpointer user_data)
+{
+	HTMLEngineSaveState *state = (HTMLEngineSaveState *) user_data;
+	const gchar *str;
+
+	str = html_engine_get_class_data (state->engine, state->save_data_class_name, g_quark_to_string (key_id));
+	/* printf ("object %s %s\n", g_quark_to_string (key_id), str); */
+	if (!str) {
+		/* printf ("save %s %s -> %s\n", state->save_data_class_name, g_quark_to_string (key_id), (gchar *) data); */
+		html_engine_save_output_string (state, "<!--+GtkHTML:<DATA class=\"%s\" key=\"%s\" value=\"%s\">",
+						state->save_data_class_name, g_quark_to_string (key_id), data);
+		html_engine_set_class_data (state->engine, state->save_data_class_name, g_quark_to_string (key_id), data);
+	}
+}
+
+static void
+handle_object_data (gpointer key, gpointer value, gpointer data)
+{
+	HTMLEngineSaveState *state = (HTMLEngineSaveState *) data;
+	gchar *str;
+
+	str = html_object_get_data (HTML_OBJECT (state->save_data_object), key);
+	/* printf ("handle: %s %s %s %s\n", state->save_data_class_name, key, value, str); */
+	if (!str) {
+		/* printf ("clear\n"); */
+		html_engine_save_output_string (state, "<!--+GtkHTML:<DATA class=\"%s\" clear=\"%s\">",
+						state->save_data_class_name, key);
+		
+		state->data_to_remove = g_slist_prepend (state->data_to_remove, key);
+	} else if (strcmp (value, str)) {
+		/* printf ("change\n"); */
+		html_engine_save_output_string (state, "<!--+GtkHTML:<DATA class=\"%s\" key=\"%s\" value=\"%s\">",
+						state->save_data_class_name, key, str);
+		html_engine_set_class_data (state->engine, state->save_data_class_name, key, value);
+	}
+}
+
+static void
+clear_data (gchar *key, HTMLEngineSaveState *state)
+{
+	html_engine_clear_class_data (state->engine, state->save_data_class_name, key);
+}
+
+gboolean
+html_object_save_data (HTMLObject *self, HTMLEngineSaveState *state)
+{
+	if (state->engine->save_data) {
+		GHashTable *t;
+		state->save_data_class_name = html_type_name (self->klass->type);
+		state->save_data_object = self;
+		t = html_engine_get_class_table (state->engine, state->save_data_class_name);
+		if (t) {
+			state->data_to_remove = NULL;
+			g_hash_table_foreach (t, handle_object_data, state);
+			g_slist_foreach (state->data_to_remove, (GFunc) clear_data, state);
+			g_slist_free (state->data_to_remove);
+			state->data_to_remove = NULL;
+		}
+		g_datalist_foreach (&self->object_data, object_save_data, state);
+	}
+
+	return TRUE;
+}
+
 GList *
 html_object_get_bound_list (HTMLObject *self, GList *list)
 {
@@ -1615,7 +1734,7 @@ unselect_object (HTMLObject *o, HTMLEngine *e, gpointer data)
 }
 
 gchar *
-html_object_get_selection_string (HTMLObject *o)
+html_object_get_selection_string (HTMLObject *o, HTMLEngine *e)
 {
 	HTMLObject *tail;
 	tmpSelData data;
@@ -1628,9 +1747,9 @@ html_object_get_selection_string (HTMLObject *o)
 	data.in     = FALSE;
 	data.i      = html_interval_new (html_object_get_head_leaf (o), tail, 0, html_object_get_length (tail));
 
-	html_interval_forall (data.i, NULL, select_object, &data);
+	html_interval_forall (data.i, e, select_object, &data);
 	html_object_append_selection_string (o, data.buffer);
-	html_interval_forall (data.i, NULL, unselect_object, NULL);
+	html_interval_forall (data.i, e, unselect_object, NULL);
 
 	html_interval_destroy (data.i);
 	string = data.buffer->str;
@@ -1745,7 +1864,7 @@ merge_down (HTMLEngine *e, GList *left, GList *right)
 		ro    = HTML_OBJECT (right->data);
 		left  = left->next;
 		right = right->next;
-		if (!html_object_merge (lo, ro, e, left, right))
+		if (!html_object_merge (lo, ro, e, &left, &right, NULL))
 			break;
 	}
 }
@@ -1755,3 +1874,141 @@ html_object_merge_down (HTMLObject *o, HTMLObject *w, HTMLEngine *e)
 {
 	merge_down (e, html_object_tails_list (o), html_object_heads_list (w));
 }
+
+gboolean
+html_object_is_parent (HTMLObject *parent, HTMLObject *child)
+{
+	g_assert (parent && child);
+
+	while (child) {
+		if (child->parent == parent)
+			return TRUE;
+		child = child->parent;
+	}
+
+	return FALSE;
+}
+
+gint
+html_object_get_insert_level (HTMLObject *o)
+{
+	switch (HTML_OBJECT_TYPE (o)) {
+	case HTML_TYPE_TABLECELL:
+	case HTML_TYPE_CLUEV:
+		return 3;
+	case HTML_TYPE_CLUEFLOW:
+		return 2;
+	default:
+		return 1;
+	}
+}
+
+void
+html_object_engine_translation (HTMLObject *o, HTMLEngine *e, gint *tx, gint *ty)
+{
+	HTMLObject *p;
+
+	*tx = 0;
+	*ty = 0;
+
+	for (p = o->parent; p != NULL && HTML_OBJECT_TYPE (p) != HTML_TYPE_IFRAME; p = p->parent) {
+		*tx += p->x;
+		*ty += p->y - p->ascent;
+	}
+
+	*tx = *tx + e->leftBorder - e->x_offset;
+	*ty = *ty + e->topBorder - e->y_offset;
+
+}
+
+gboolean
+html_object_engine_intersection (HTMLObject *o, HTMLEngine *e, gint tx, gint ty, gint *x1, gint *y1, gint *x2, gint *y2)
+{
+	*x1 = o->x + tx;
+	*y1 = o->y - o->ascent + ty;
+	*x2 = o->x + o->width + tx;
+	*y2 = o->y + o->descent + ty;
+
+	return html_engine_intersection (e, x1, y1, x2, y2);
+}
+
+void
+html_object_add_to_changed (GList **changed_objs, HTMLObject *o)
+{
+	GList *l, *next;
+
+	if (!changed_objs || (*changed_objs && (*changed_objs)->data == o))
+		return;
+
+	for (l = *changed_objs; l; l = next) {
+		if (l->data == NULL) {
+			l = l->next;
+			next = l->next;
+			continue;
+		}
+		next = l->next;
+		if (html_object_is_parent (o, HTML_OBJECT (l->data))) {
+			*changed_objs = g_list_remove_link (*changed_objs, l);
+			g_list_free (l);
+		} else
+			break;
+	}
+
+	*changed_objs = g_list_prepend (*changed_objs, o);
+}
+
+HTMLClearType
+html_object_get_clear (HTMLObject *self)
+{
+	return (* HO_CLASS (self)->get_clear) (self);
+}
+
+static HTMLObject *
+next_prev_cursor_object (HTMLObject *o, HTMLEngine *e, gint *offset, gboolean forward)
+{
+	HTMLCursor cursor;
+	gboolean result;
+
+	html_cursor_init (&cursor, o, HTML_IS_TABLE (o) ? *offset : (forward ? html_object_get_length (o) : 0));
+
+	result = forward ? html_cursor_forward (&cursor, e) : html_cursor_backward (&cursor, e);
+	*offset = cursor.offset;
+
+	return result ? cursor.object : NULL;
+}
+
+HTMLObject *
+html_object_next_cursor_object (HTMLObject *o, HTMLEngine *e, gint *offset)
+{
+	return next_prev_cursor_object (o, e, offset, TRUE);
+}
+
+HTMLObject *
+html_object_prev_cursor_object (HTMLObject *o, HTMLEngine *e, gint *offset)
+{
+	return next_prev_cursor_object (o, e, offset, FALSE);
+}
+
+HTMLObject *
+html_object_next_cursor_leaf (HTMLObject *o, HTMLEngine *e)
+{
+	gint offset = html_object_get_length (o);
+
+	o = html_object_next_cursor_object (o, e, &offset);
+	while (o && html_object_is_container (o))
+		o = html_object_next_cursor_object (o, e, &offset);
+
+	return o;
+}
+
+HTMLObject *
+html_object_prev_cursor_leaf (HTMLObject *o, HTMLEngine *e)
+{
+	gint offset = html_object_get_length (o);
+
+	o = html_object_prev_cursor_object (o, e, &offset);
+	while (o && html_object_is_container (o))
+		o = html_object_prev_cursor_object (o, e, &offset);
+
+	return o;
+}
